#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A generic unit representatives used for testing.."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
from thoth.python import PackageVersion

from thoth.adviser.state import State
from thoth.adviser.step import Step
from voluptuous import Required
from voluptuous import Schema

if TYPE_CHECKING:
    from thoth.adviser.pipeline_builder import PipelineBuilderContext


class Step1(Step):
    """A testing step implementation."""

    CONFIGURATION_DEFAULT = {"guido_retirement": 2019, "package_name": "tensorflow", "multi_package_resolution": False}
    CONFIGURATION_SCHEMA: Schema = Schema(
        {Required("package_name"): str, Required("guido_retirement"): int, Required("multi_package_resolution"): bool}
    )

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Run noop method."""


class Step2(Step):
    """A testing step implementation."""

    CONFIGURATION_DEFAULT = {
        "lyrics": ".. drifting further everyday...",
        "package_name": "thoth-adviser",
        "multi_package_resolution": False,
    }
    CONFIGURATION_SCHEMA: Schema = Schema(
        {Required("package_name"): str, Required("lyrics"): str, Required("multi_package_resolution"): bool}
    )

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if this pipeline unit should be included in the pipeline configuration."""
        yield from ()
        return None

    def run(
        self, state: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Run noop method."""


__all__ = ["Step1", "Step2"]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A wrap that links Python package version to Thoth Search UI."""

import os
from typing import TYPE_CHECKING
from typing import Generator
from typing import Dict
from typing import Any
from urllib.parse import quote

import attr
from ..state import State
from ..wrap import Wrap

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_THOTH_SEARCH_PACKAGE_URL = os.getenv(
    "THOTH_SEARCH_PACKAGE_URL",
    "https://thoth-station.ninja/search/package/{package_name}/"
    "{package_version}/{index_url}/{os_name}/{os_version}/{python_version}",
)


@attr.s(slots=True)
class ThothSearchPackageWrap(Wrap):
    """A wrap that links Python package to Thoth Search UI."""

    _search_package_url = attr.ib(type=str, default=_THOTH_SEARCH_PACKAGE_URL, kw_only=True)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[Any, Any], None, None]:
        """Include this wrap in adviser, once."""
        if not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self, state: State) -> None:
        """Add a link to Thoth Search UI for each resolved package."""
        runtime_environment = self.context.project.runtime_environment
        for package_version_tuple in state.resolved_dependencies.values():
            state.justification.append(
                {
                    "type": "INFO",
                    "link": self._search_package_url.format(
                        package_name=quote(package_version_tuple[0]),
                        package_version=quote(package_version_tuple[1], safe=""),
                        index_url=quote(package_version_tuple[2], safe=""),
                        os_name=quote(runtime_environment.operating_system.name),
                        os_version=quote(runtime_environment.operating_system.version),
                        python_version=quote(runtime_environment.python_version),
                    ),
                    "message": f"Browse Thoth Search UI for '{package_version_tuple[0]}=={package_version_tuple[1]}'",
                    "package_name": package_version_tuple[0],
                }
            )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to provide information about the last CVE update."""

import pytest

from datetime import datetime
from datetime import timedelta

from thoth.adviser.boots import CveTimestampBoot
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.enums import RecommendationType
from thoth.common import datetime2datetime_str
from thoth.common import get_justification_link as jl
from ..base import AdviserUnitTestCase


class TestCveTimestampBoot(AdviserUnitTestCase):
    """Test providing information about the last CVE update."""

    UNIT_TESTED = CveTimestampBoot

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test registering this unit for adviser runs."""
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

        builder_context.add_unit(self.UNIT_TESTED())

        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    @pytest.mark.parametrize(
        "timestamp,type_severity",
        [(datetime.utcnow() - timedelta(days=1), "INFO"), (datetime.utcnow() - timedelta(days=1000), "WARNING")],
    )
    def test_run_info(self, context: Context, timestamp: datetime, type_severity: str) -> None:
        """Test providing CVE update timestamp to users."""
        context.graph.should_receive("get_cve_timestamp").and_return(timestamp).once()
        unit = self.UNIT_TESTED()

        assert not context.stack_info
        with unit.assigned_context(context):
            unit.run()

        assert len(context.stack_info) == 1
        assert set(context.stack_info[0]) == {"link", "message", "type"}
        assert self.verify_justification_schema(context.stack_info)
        assert context.stack_info[0]["link"] == jl("cve_timestamp")
        assert (
            context.stack_info[0]["message"] == f"CVE database of known vulnerabilities for Python packages "
            f"was updated at {datetime2datetime_str(timestamp)!r}"
        )
        assert context.stack_info[0]["type"] == type_severity

    def test_run_no_cve_timestamp(self, context: Context) -> None:
        """Test warning if CVE timestamp is not recorded in the database."""
        context.graph.should_receive("get_cve_timestamp").and_return(None).once()
        unit = self.UNIT_TESTED()

        assert not context.stack_info
        with unit.assigned_context(context):
            unit.run()

        assert len(context.stack_info) == 1
        assert set(context.stack_info[0]) == {"link", "message", "type"}
        assert self.verify_justification_schema(context.stack_info)
        assert context.stack_info[0] == {
            "link": jl("no_cve_timestamp"),
            "message": "No CVE timestamp information found in the database",
            "type": "WARNING",
        }



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test wrap adding information about Red Hat's Pulp instance."""

from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.state import State
from thoth.adviser.wraps import PulpReleaseWrap

from ..base import AdviserUnitTestCase


class TestPyPIReleaseWrap(AdviserUnitTestCase):
    """Test wrap adding information about Red Hat's Pulp instance."""

    UNIT_TESTED = PulpReleaseWrap

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_run_add_justification(self) -> None:
        """Test adding a link to Pulp instance for the given package released."""
        state = State()
        package_version_tuple = ("tensorflow", "2.5.0", f"{self.UNIT_TESTED._PULP_URL}pypi/test/simple")
        state.add_resolved_dependency(package_version_tuple)

        unit = self.UNIT_TESTED()
        unit.run(state)

        assert state.justification == [
            {
                "type": "INFO",
                "message": f"Package {package_version_tuple[0]!r} is released on Operate First Pulp instance",
                "package_name": package_version_tuple[0],
                "link": package_version_tuple[2],
            }
        ]
        self.verify_justification_schema(state.justification)

    def test_run_no_justification(self) -> None:
        """Test NOT adding a link to Pulp instance for the given package released."""
        index_url = "https://pypi.org/simple"
        assert not self.UNIT_TESTED._PULP_URL.startswith(index_url)

        state = State()
        state.add_resolved_dependency(("tensorflow", "2.5.0", index_url))

        unit = self.UNIT_TESTED()
        unit.run(state)

        assert not state.justification



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of Adaptive Simulated Annealing (ASA) used to resolve software stacks."""

from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import TYPE_CHECKING
import logging
import random
import math

import attr

from ..context import Context
from ..exceptions import NoHistoryKept
from ..predictor import Predictor
from ..state import State

if TYPE_CHECKING:
    import matplotlib

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class AdaptiveSimulatedAnnealing(Predictor):
    """Implementation of adaptive simulated annealing looking for stacks based on the scoring function."""

    temperature_coefficient = attr.ib(type=float, default=0.999, kw_only=True)

    _temperature_history = attr.ib(
        type=List[Tuple[Optional[float], Optional[bool], Optional[float], int]],
        factory=list,
        kw_only=True,
    )
    _temperature = attr.ib(type=float, kw_only=True, default=0.0)

    def _temperature_function(self, t0: float, context: Context) -> float:
        """Temperature function used to compute new temperature."""
        k = (context.accepted_final_states_count + math.log(context.iteration + 1)) / context.limit
        temperature = t0 * self.temperature_coefficient**k
        _LOGGER.debug(
            "New temperature for (iteration=%d, t0=%g, accepted final states=%d, limit=%d, beam size= %d, k=%f) = %g",
            context.iteration,
            t0,
            context.accepted_final_states_count,
            context.limit,
            context.beam.size,
            k,
            temperature,
        )
        return max(temperature, 0.0)

    @staticmethod
    def _compute_acceptance_probability(top_score: float, neighbour_score: float, temperature: float) -> float:
        """Check the probability of acceptance the given solution to expansion."""
        if temperature == 0.0:
            # This can happen as we consider also iteration in the exp function - if we
            # drop bellow 0, acceptance probability drops to 0.
            return 0.0

        if neighbour_score > top_score:
            return 1.0

        acceptance_probability = math.exp((neighbour_score - top_score) / temperature)
        _LOGGER.debug(
            "Acceptance probability for (top_score=%g, neighbour_score=%g, temperature=%g) = %g",
            top_score,
            neighbour_score,
            temperature,
            acceptance_probability,
        )
        return acceptance_probability

    def run(self) -> Tuple[State, Tuple[str, str, str]]:
        """Run adaptive simulated annealing on top of beam."""
        self._temperature = self._temperature_function(self._temperature, self.context)

        # Expand highest promising by default.
        state = self.context.beam.max()

        # Pick a random state to be expanded if accepted.
        probable_state_idx = random.randrange(1, self.context.beam.size) if self.context.beam.size > 1 else 0
        probable_state = self.context.beam.get(probable_state_idx)
        acceptance_probability = self._compute_acceptance_probability(
            state.score, probable_state.score, self._temperature
        )

        if probable_state_idx != 0 and acceptance_probability >= random.random():
            # Skip to probable state, do not use the top rated state.
            _LOGGER.debug(
                "Performing transition to a neighbour state with score %g",
                probable_state.score,
            )
            # state_expansion_idx = probable_state_idx
            state = probable_state
            unresolved_dependency_tuple = state.get_random_unresolved_dependency(prefer_recent=False)
        else:
            _LOGGER.debug("Expanding TOP rated state with score %g", state.score)
            unresolved_dependency_tuple = state.get_random_unresolved_dependency(prefer_recent=True)

        if self.keep_history:
            self._temperature_history.append(
                (
                    self._temperature,
                    state is self.context.beam.max(),
                    acceptance_probability,
                    self.context.accepted_final_states_count,
                )
            )

        return state, unresolved_dependency_tuple

    def pre_run(self) -> None:
        """Initialize before the actual annealing run."""
        self._temperature_history.clear()
        self._temperature = float(self.context.limit)

    @staticmethod
    def _make_patch_spines_invisible(ax: Any) -> None:
        """Make spines invisible."""
        ax.set_frame_on(True)
        ax.patch.set_visible(False)
        for sp in ax.spines.values():
            sp.set_visible(False)

    def plot(self) -> "matplotlib.figure.Figure":
        """Plot temperature history of adaptive simulated annealing."""
        # Code adjusted based on:
        #    https://matplotlib.org/3.1.1/gallery/ticks_and_spines/multiple_yaxis_with_spines.html
        if self._temperature_history is None:
            raise NoHistoryKept("No history datapoints kept")

        import matplotlib.pyplot as plt
        from matplotlib.font_manager import FontProperties

        x = [i for i in range(len(self._temperature_history))]
        # Top rated candidate was chosen.
        y1 = [i[0] if i[1] is True else None for i in self._temperature_history]
        # A neighbour candidate was chosen.
        y2 = [i[0] if i[1] is False else None for i in self._temperature_history]
        # Acceptance probability - as the probability in 0 - 1, lets make it larger - scale to temperature size.
        y3 = [i[2] for i in self._temperature_history]
        # Number of products.
        y4 = [i[3] for i in self._temperature_history]

        fig, host = plt.subplots()
        fig.subplots_adjust(right=0.75)

        par1 = host.twinx()
        par2 = host.twinx()

        # Offset the right spine of par1 and par2. The ticks and label have already been
        # placed on the right by twinx above.
        par1.spines["right"].set_position(("axes", 1.050))
        par2.spines["right"].set_position(("axes", 1.225))

        # Having been created by twinx, par1 par2 have their frame off, so the line of its
        # detached spine is invisible. First, activate the frame but make the patch
        # and spines invisible.
        self._make_patch_spines_invisible(par1)
        self._make_patch_spines_invisible(par2)

        # Second, show the right spine.
        par1.spines["right"].set_visible(True)
        par2.spines["right"].set_visible(True)
        host.spines["right"].set_visible(False)
        host.spines["top"].set_visible(False)

        host.plot(x, y1, ".g", label="Expansion of a highest rated candidate")
        host.plot(x, y2, ",r", label="Expansion of a neighbour candidate")
        (p3,) = par1.plot(x, y3, ",b", label="Acceptance probability for a neighbour candidate")
        (p4,) = par2.plot(x, y4, ",y", label="Number of products produced in the pipeline")

        host.set_xlabel("iteration")
        host.set_ylabel("temperature")
        par1.set_ylabel("acceptance probability")
        par2.set_ylabel("product count")

        host.yaxis.label.set_color("black")
        par1.yaxis.label.set_color(p3.get_color())
        par2.yaxis.label.set_color(p4.get_color())

        tkw = dict(size=4, width=1.5)
        host.tick_params(axis="y", colors="black", **tkw)
        par1.tick_params(axis="y", colors=p3.get_color(), **tkw)
        par2.tick_params(axis="y", colors=p4.get_color(), **tkw)
        host.tick_params(axis="x", **tkw)

        font_prop = FontProperties()
        font_prop.set_size("small")
        fig.legend(
            loc="upper center",
            bbox_to_anchor=(0.50, 1.00),
            ncol=2,
            fancybox=True,
            shadow=True,
            prop=font_prop,
        )
        return fig



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A base class for implementing pipeline configuration."""

from itertools import chain
from typing import Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Union

import attr

from .boot import Boot
from .dm_report import DependencyMonkeyReport
from .exceptions import PipelineUnitError
from .pseudonym import Pseudonym
from .sieve import Sieve
from .step import Step
from .stride import Stride
from .wrap import Wrap

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .unit import Unit
    from .report import Report


@attr.s(slots=True)
class PipelineConfig:
    """A configuration of a pipeline for dependency-monkey and for adviser."""

    _boots = attr.ib(type=Dict[Optional[str], List[Boot]], factory=dict)
    # Use a dict to have O(1) access when applying pseudonyms.
    _pseudonyms = attr.ib(type=Dict[str, List[Pseudonym]], factory=dict)
    _sieves = attr.ib(type=Dict[Optional[str], List[Sieve]], factory=dict)
    _steps = attr.ib(type=Dict[Optional[str], List[Step]], factory=dict)
    _strides = attr.ib(type=Dict[Optional[str], List[Stride]], factory=dict)
    _wraps = attr.ib(type=Dict[Optional[str], List[Wrap]], factory=dict)

    @property
    def boots(self) -> List[Boot]:
        """Get all boots."""
        return list(chain(*self._boots.values()))

    @property
    def boots_dict(self) -> Dict[Optional[str], List[Boot]]:
        """Get boots as a dictionary mapping."""
        return self._boots

    @property
    def pseudonyms(self) -> List[Pseudonym]:
        """Get all pseudonyms."""
        return list(chain(*self._pseudonyms.values()))

    @property
    def pseudonyms_dict(self) -> Dict[str, List[Pseudonym]]:
        """Get pseudonyms as a dictionary mapping."""
        return self._pseudonyms

    @property
    def sieves(self) -> List[Sieve]:
        """Get all sieves."""
        return list(chain(*self._sieves.values()))

    @property
    def sieves_dict(self) -> Dict[Optional[str], List[Sieve]]:
        """Get sieves as a dictionary mapping."""
        return self._sieves

    @property
    def steps(self) -> List[Step]:
        """Get all steps."""
        return list(chain(*self._steps.values()))

    @property
    def steps_dict(self) -> Dict[Optional[str], List[Step]]:
        """Get steps as a dictionary mapping."""
        return self._steps

    @property
    def strides(self) -> List[Stride]:
        """Get all strides."""
        return list(chain(*self._strides.values()))

    @property
    def strides_dict(self) -> Dict[Optional[str], List[Stride]]:
        """Get strides as a dictionary mapping."""
        return self._strides

    @property
    def wraps(self) -> List[Wrap]:
        """Get all wraps."""
        return list(chain(*self._wraps.values()))

    @property
    def wraps_dict(self) -> Dict[Optional[str], List[Wrap]]:
        """Get wraps as a dictionary mapping."""
        return self._wraps

    def to_dict(self) -> Dict[str, List[Dict[str, Any]]]:
        """Return this pipeline configuration in a dict representation."""
        return {
            "boots": [boot.to_dict() for boot in self.boots],
            "pseudonyms": [pseudonym.to_dict() for pseudonym in self.pseudonyms],
            "sieves": [sieve.to_dict() for sieve in self.sieves],
            "steps": [step.to_dict() for step in self.steps],
            "strides": [stride.to_dict() for stride in self.strides],
            "wraps": [wrap.to_dict() for wrap in self.wraps],
        }

    def iter_units(self):
        # type:('PipelineConfig') -> Generator[Unit, None, None]
        """Iterate over units present in the configuration."""
        yield from chain(self.boots, self.pseudonyms, self.sieves, self.steps, self.strides, self.wraps)

    def iter_units_reversed(self):
        # type:('PipelineConfig') -> Generator[Unit, None, None]
        """Iterate over units present in the configuration in a reversed order."""
        yield from reversed(self.wraps)
        yield from reversed(self.strides)
        yield from reversed(self.steps)
        yield from reversed(self.sieves)
        yield from reversed(self.pseudonyms)
        yield from reversed(self.boots)

    def call_pre_run(self) -> None:
        """Call pre-run method on all units registered in this configuration."""
        for unit in self.iter_units():
            try:
                unit.pre_run()
            except Exception as exc:
                raise PipelineUnitError(f"Failed to run pre_run method on unit {unit.name!r}: {str(exc)}") from exc

    def call_post_run(self) -> None:
        """Call post-run method on all units registered in this configuration."""
        for unit in self.iter_units_reversed():
            try:
                unit.post_run()
            except Exception as exc:
                raise PipelineUnitError(f"Failed to run post_run method on unit {unit.name!r}: {str(exc)}") from exc

    def call_post_run_report(self, report):
        # type:('PipelineConfig', Union[Report, DependencyMonkeyReport]) -> None
        """Call post-run method when report is generated."""
        for unit in self.iter_units_reversed():
            try:
                unit.post_run_report(report)
            except Exception as exc:
                raise PipelineUnitError(
                    f"Failed to run pre_run_report method on unit {unit.name!r}: {str(exc)}"
                ) from exc



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out packages based on index configuration."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from voluptuous import Schema
from voluptuous import Required

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class PackageIndexConfigurationSieve(Sieve):
    """A sieve to filter out packages based on index configuration."""

    CONFIGURATION_DEFAULT = {"package_name": None, "index_url": None}
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): str, Required("index_url"): str})
    _JUSTIFICATION_LINK = jl("index_package")

    packages_seen = attr.ib(type=Set[Tuple[str, str, str]], default=attr.Factory(set), init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register for each package."""
        if not builder_context.is_included(cls) and builder_context.project.pipfile.thoth.disable_index_adjustment:
            for package_version in builder_context.project.iter_dependencies(with_devel=True):
                if package_version.index:
                    yield {
                        "package_name": package_version.name,
                        "index_url": package_version.index.url,
                    }

            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        self.packages_seen.clear()
        super().pre_run()

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Cut-off pre-releases if project does not explicitly allows them."""
        for package_version in package_versions:
            if package_version.index.url != self.configuration["index_url"]:
                package_tuple = package_version.to_tuple()

                if package_tuple not in self.packages_seen:
                    self.packages_seen.add(package_tuple)
                    msg = (
                        f"Skipping package {package_tuple} as it does not used configured "
                        f"index {self.configuration['index_url']!r}"
                    )
                    _LOGGER.warning("%s - see %s", msg, self._JUSTIFICATION_LINK)
                    self.context.stack_info.append(
                        {
                            "message": msg,
                            "type": "WARNING",
                            "link": self._JUSTIFICATION_LINK,
                        }
                    )

                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test handling checks of supported platforms."""

import pytest

from thoth.adviser.boots import PlatformBoot
from thoth.adviser.context import Context
from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.common import get_justification_link as jl
from ..base import AdviserUnitTestCase


class TestPlatformBoot(AdviserUnitTestCase):
    """Test platform boot."""

    UNIT_TESTED = PlatformBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context.project.runtime_environment.platform = "aarch64"
        self.verify_multiple_should_include(builder_context)

    def test_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test registering this unit."""
        builder_context.should_receive("is_included").and_return(False)
        builder_context.project.runtime_environment.platform = "aarch64"
        assert list(self.UNIT_TESTED.should_include(builder_context)) == [{}]

        builder_context.should_receive("is_included").and_return(True)
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_not_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test not including the pipeline unit when supported platforms are used."""
        builder_context.should_receive("is_included").and_return(False)
        builder_context.project.runtime_environment.platform = "linux-x86_64"
        assert list(self.UNIT_TESTED.should_include(builder_context)) == []

    def test_not_acceptable(self, context: Context) -> None:
        """Test raising an exception when the pipeline unit is included."""
        unit = self.UNIT_TESTED()

        context.project.runtime_environment.platform = "aarch64"

        assert not context.stack_info

        # The unit always raises, no need to explicitly assign platform here.
        with unit.assigned_context(context):
            with pytest.raises(
                NotAcceptable, match="Platform 'aarch64' is not supported, possible platforms are: linux-x86_64"
            ):
                unit.run()

        assert self.verify_justification_schema(context.stack_info)
        assert context.stack_info == [
            {
                "link": jl("platform"),
                "message": "Platform 'aarch64' is not supported, possible platforms are: " "linux-x86_64",
                "type": "ERROR",
            },
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to notify about labels used during the resolution."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class LabelsBoot(Boot):
    """A boot to notify about labels used during the resolution."""

    _JUSTIFICATION_LINK_LABELS = jl("labels")
    _JUSTIFICATION_LINK_ALLOW_CVE = jl("allow_cve")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self if labels are used."""
        if not builder_context.is_included(cls) and builder_context.labels:
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Notify about labels used during the resolution process.."""
        for key, value in self.context.labels.items():
            if key != "allow-cve":
                msg = f"Considering label {key}={value} in the resolution process"
                _LOGGER.info("%s - see %s", msg, self._JUSTIFICATION_LINK_LABELS)
                self.context.stack_info.append(
                    {
                        "message": msg,
                        "type": "INFO",
                        "link": self._JUSTIFICATION_LINK_LABELS,
                    }
                )
            else:
                for allow_cve in value.split(","):
                    msg = f"Allowing CVE {allow_cve.upper()!r} to be present in the application"
                    _LOGGER.warning("%s - see %s", msg, self._JUSTIFICATION_LINK_ALLOW_CVE)
                    self.context.stack_info.append(
                        {
                            "message": msg,
                            "type": "WARNING",
                            "link": self._JUSTIFICATION_LINK_ALLOW_CVE,
                        }
                    )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""JSON schema for pipeline unit prescription in version v1."""

import re
from urllib.parse import urlparse

from voluptuous import All
from voluptuous import Any
from voluptuous import Invalid
from voluptuous import Length
from voluptuous import Optional
from voluptuous import Range
from voluptuous import Required
from voluptuous import Schema

from thoth.python import PackageVersion
from packaging.specifiers import SpecifierSet
from packaging.specifiers import InvalidSpecifier

from thoth.adviser.cpu_db import CPUDatabase
from thoth.adviser.enums import DecisionType
from thoth.adviser.enums import RecommendationType


_NONEMPTY_STRING = All(str, Length(min=1))
_NONEMPTY_LIST_OF_NONEMPTY_STRINGS = All([_NONEMPTY_STRING], Length(min=1))
_NONEMPTY_LIST_OF_NONEMPTY_STRINGS_WITH_NONE = All([_NONEMPTY_STRING, None], Length(min=1))
_NONEMPTY_LIST_OF_INTEGERS_WITH_NONE = All([int, None], Length(min=1))


def _with_not(entity: object) -> Schema:
    """Add possibility to provide 'not' in the given configuration entity."""
    return Schema(Any(entity, Schema({"not": entity})))


def _specifier_set(v: object) -> None:
    """Validate a specifier set."""
    if not isinstance(v, str):
        raise Invalid(f"Value {v!r} is not valid version specifier (example: '<1.0,>=0.5')")
    try:
        SpecifierSet(v)
    except InvalidSpecifier as exc:
        raise Invalid(str(exc))


def _python_package_name(n: object) -> None:
    """Validate Python package name."""
    if not isinstance(n, str):
        raise Invalid(f"Value {n!r} is not a valid package name")

    try:
        normalized = PackageVersion.normalize_python_package_name(n)
    except Exception as exc:
        raise Invalid(f"Failed to parse Python package name {n!r}: {str(exc)}")
    else:
        if normalized != n:
            raise Invalid(f"Python package name {n!r} is not in a normalized form, normalized: {normalized!r}")


_RPM_PACKAGE_VERSION_SCHEMA = Schema(
    {
        Required("package_name"): _NONEMPTY_STRING,
        Optional("arch"): _NONEMPTY_STRING,
        Optional("epoch"): _NONEMPTY_STRING,
        Optional("package_identifier"): _NONEMPTY_STRING,
        Optional("package_version"): _NONEMPTY_STRING,
        Optional("release"): _NONEMPTY_STRING,
        Optional("src"): bool,
    }
)

_PYTHON_PACKAGE_VERSION_SCHEMA = Schema(
    {
        Required("name"): _python_package_name,
        Optional("version"): _specifier_set,
        Optional("location"): Any(_NONEMPTY_STRING, None),
    }
)

PRESCRIPTION_UNIT_SHOULD_INCLUDE_RUNTIME_ENVIRONMENTS_SCHEMA = Schema(
    {
        Optional("operating_systems"): [
            Schema(
                {
                    Optional("name"): _NONEMPTY_STRING,
                    Optional("version"): _NONEMPTY_STRING,
                }
            )
        ],
        Optional("hardware"): All(
            [
                Schema(
                    {
                        Optional("cpu_families"): _with_not(_NONEMPTY_LIST_OF_INTEGERS_WITH_NONE),
                        Optional("cpu_models"): _with_not(_NONEMPTY_LIST_OF_INTEGERS_WITH_NONE),
                        Optional("cpu_flags"): _with_not(All(CPUDatabase.get_known_flags(), Length(min=1))),
                        Optional("gpu_models"): _with_not(_NONEMPTY_LIST_OF_NONEMPTY_STRINGS_WITH_NONE),
                    }
                )
            ],
            Length(min=1),
        ),
        Optional("python_version"): Any(_specifier_set, None),
        Optional("cuda_version"): Any(_specifier_set, None),
        Optional("platforms"): _with_not(_NONEMPTY_LIST_OF_NONEMPTY_STRINGS_WITH_NONE),
        Optional("openblas_version"): Any(_specifier_set, None),
        Optional("openmpi_version"): Any(_specifier_set, None),
        Optional("cudnn_version"): Any(_specifier_set, None),
        Optional("mkl_version"): Any(_specifier_set, None),
        Optional("base_images"): _with_not(_NONEMPTY_LIST_OF_NONEMPTY_STRINGS_WITH_NONE),
        Optional("abi"): _with_not(_NONEMPTY_LIST_OF_NONEMPTY_STRINGS),
        Optional("rpm_packages"): _with_not(All([_RPM_PACKAGE_VERSION_SCHEMA], Length(min=1))),
        Optional("python_packages"): _with_not(All([_PYTHON_PACKAGE_VERSION_SCHEMA], Length(min=1))),
    }
)


def _library_usage(v: object) -> None:
    if not isinstance(v, dict):
        raise Invalid(f"Expected a dictionary describing library usage, got: {v}")

    if not v:
        raise Invalid(f"Got empty library usage: {v}")

    for k, v in v.items():
        if not isinstance(k, str):
            raise Invalid(f"Unknown key representing package name in library usage: {k}")

        if not k:
            raise Invalid("Empty key representing package name in library usage")

        # A list of library calls.
        _NONEMPTY_LIST_OF_NONEMPTY_STRINGS(v)


def _labels(v: object) -> None:
    if not isinstance(v, dict):
        raise Invalid(f"Expected a dictionary describing labels, got {v}")

    for val in v.values():
        if not isinstance(val, str):
            raise Invalid(f"Expected a non-empty string for label value, got {type(val)} instead: {val!r}")

        _NONEMPTY_STRING(val)


PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA = Schema(
    {
        Optional("times", default=1): All(int, Range(min=0, max=1)),
        Optional("adviser_pipeline"): bool,
        Optional("dependency_monkey_pipeline"): bool,
        Optional("labels"): _labels,
        Optional("dependencies"): Schema(
            {
                Optional("boots"): _NONEMPTY_LIST_OF_NONEMPTY_STRINGS,
                Optional("pseudonyms"): _NONEMPTY_LIST_OF_NONEMPTY_STRINGS,
                Optional("sieves"): _NONEMPTY_LIST_OF_NONEMPTY_STRINGS,
                Optional("steps"): _NONEMPTY_LIST_OF_NONEMPTY_STRINGS,
                Optional("strides"): _NONEMPTY_LIST_OF_NONEMPTY_STRINGS,
                Optional("wraps"): _NONEMPTY_LIST_OF_NONEMPTY_STRINGS,
            }
        ),
        Optional("recommendation_types"): _with_not(
            All(list(map(str.lower, RecommendationType.__members__.keys())), Length(min=1))
        ),
        Optional("decision_types"): _with_not(
            All(list(map(str.lower, DecisionType.__members__.keys())), Length(min=1))
        ),
        Optional("runtime_environments"): PRESCRIPTION_UNIT_SHOULD_INCLUDE_RUNTIME_ENVIRONMENTS_SCHEMA,
        Optional("library_usage"): _library_usage,
        Optional("authenticated"): bool,
    }
)


_METADATA_SCHEMA = Schema({Required(str): object})


_UNIT_SCHEMA_BASE_DICT = {
    Required("name"): _NONEMPTY_STRING,
    Required("should_include"): PRESCRIPTION_UNIT_SHOULD_INCLUDE_SCHEMA,
    Optional("metadata"): _METADATA_SCHEMA,
}


def _justification_link(v: str) -> None:
    """Validate justification link."""
    if v.startswith(("https://", "http://")):
        try:
            urlparse(v)
        except Exception as exc:
            raise Invalid(f"Failed to validate URL: {str(exc)}")
    else:
        matched = re.fullmatch(r"[a-z0-9_-]+", v)
        if not matched:
            raise Invalid(f"Failed to validate base justification link {v!r}")

    return None


STACK_INFO_SCHEMA = Schema(
    {
        Required("type"): Any("WARNING", "INFO", "ERROR"),
        Required("message"): _NONEMPTY_STRING,
        Required("link"): _justification_link,
    }
)


JUSTIFICATION_SCHEMA = Schema(
    {
        Required("type"): Any("WARNING", "INFO", "ERROR"),
        Required("message"): _NONEMPTY_STRING,
        Required("link"): _justification_link,
        Optional("package_name"): _python_package_name,
        Optional("advisory"): _NONEMPTY_STRING,
        Optional("cve_id"): _NONEMPTY_STRING,
        Optional("cve_name"): _NONEMPTY_STRING,
        Optional("version_range"): _specifier_set,
    }
)


_UNIT_RUN_SCHEMA_BASE_DICT = {
    Optional("stack_info"): [STACK_INFO_SCHEMA],
    Optional("log"): Schema(
        {
            Required("message"): _NONEMPTY_STRING,
            Required("type"): Any("WARNING", "INFO", "ERROR"),
        }
    ),
}


def _locked_version(v: object) -> None:
    """Validate locked version."""
    if v is not None and (not isinstance(v, str) or not v.startswith("==") and not v.startswith("===")):
        raise Invalid(f"Value {v!r} is not valid locked version (example: '==1.0.0')")


PACKAGE_VERSION_LOCKED_SCHEMA = Schema(
    {
        Required("name"): Optional(_python_package_name),
        Optional("locked_version"): Optional(_locked_version),
        Optional("index_url"): Optional(_NONEMPTY_STRING),
    }
)


PACKAGE_VERSION_SCHEMA = Schema(
    {
        Optional("name"): Optional(_python_package_name),
        Optional("version"): _specifier_set,
        Optional("index_url"): Optional(Any(_NONEMPTY_STRING, {"not": _NONEMPTY_STRING})),
        Optional("develop"): bool,
    }
)


PACKAGE_VERSION_REQUIRED_NAME_SCHEMA = Schema(
    {
        Required("name"): _python_package_name,
        Optional("version"): _specifier_set,
        Optional("index_url"): Optional(Any(_NONEMPTY_STRING, {"not": _NONEMPTY_STRING})),
        Optional("develop"): bool,
    }
)

#
# Boot unit.
#

PRESCRIPTION_BOOT_MATCH_ENTRY_SCHEMA = Schema({"package_name": Optional(_NONEMPTY_STRING)})

PRESCRIPTION_BOOT_RUN_SCHEMA = Schema(
    {
        Optional("not_acceptable"): _NONEMPTY_STRING,
        Optional("eager_stop_pipeline"): _NONEMPTY_STRING,
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)

PRESCRIPTION_BOOT_SCHEMA = Schema(
    {
        Optional("match"): Any(
            All([PRESCRIPTION_BOOT_MATCH_ENTRY_SCHEMA], Length(min=1)), PRESCRIPTION_BOOT_MATCH_ENTRY_SCHEMA
        ),
        Required("run"): PRESCRIPTION_BOOT_RUN_SCHEMA,
        Required("type"): "boot",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)

#
# Pseudonym unit.
#

PRESCRIPTION_PSEUDONYM_MATCH_ENTRY_SCHEMA = Schema({"package_version": PACKAGE_VERSION_REQUIRED_NAME_SCHEMA})

PRESCRIPTION_PSEUDONYM_RUN_SCHEMA = Schema(
    {
        Required("yield"): Schema(
            {
                Optional("yield_matched_version"): bool,
                "package_version": PACKAGE_VERSION_LOCKED_SCHEMA,
            }
        ),
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)

PRESCRIPTION_PSEUDONYM_SCHEMA = Schema(
    {
        Required("match"): Any(
            All([PRESCRIPTION_PSEUDONYM_MATCH_ENTRY_SCHEMA], Length(min=1)),
            PRESCRIPTION_PSEUDONYM_MATCH_ENTRY_SCHEMA,
        ),
        Required("run"): PRESCRIPTION_PSEUDONYM_RUN_SCHEMA,
        Required("type"): "pseudonym",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)

#
# Sieve unit.
#

PRESCRIPTION_SIEVE_MATCH_ENTRY_SCHEMA = Schema({"package_version": PACKAGE_VERSION_SCHEMA})

PRESCRIPTION_SIEVE_RUN_SCHEMA = Schema(
    {
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)


PRESCRIPTION_SIEVE_SCHEMA = Schema(
    {
        Required("match"): Any(
            All([PRESCRIPTION_SIEVE_MATCH_ENTRY_SCHEMA], Length(min=1)), PRESCRIPTION_SIEVE_MATCH_ENTRY_SCHEMA
        ),
        Optional("run"): PRESCRIPTION_SIEVE_RUN_SCHEMA,
        Required("type"): "sieve",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)

#
# Skip package sieve.
#

PRESCRIPTION_SKIP_PACKAGE_SIEVE_RUN_ENTRY_SCHEMA = Schema(
    {
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)

PRESCRIPTION_SKIP_PACKAGE_SIEVE_MATCH_ENTRY_SCHEMA = Schema(
    {
        Required("package_name"): Optional(_NONEMPTY_STRING),
    }
)

PRESCRIPTION_SKIP_PACKAGE_SIEVE_SCHEMA = Schema(
    {
        Required("match"): Any(
            All([PRESCRIPTION_SKIP_PACKAGE_SIEVE_MATCH_ENTRY_SCHEMA], Length(min=1)),
            PRESCRIPTION_SKIP_PACKAGE_SIEVE_MATCH_ENTRY_SCHEMA,
        ),
        Required("type"): "sieve.SkipPackage",
        Optional("run"): PRESCRIPTION_SKIP_PACKAGE_SIEVE_RUN_ENTRY_SCHEMA,
        **_UNIT_SCHEMA_BASE_DICT,
    }
)

#
# Step unit.
#

PRESCRIPTION_STEP_MATCH_ENTRY_SCHEMA = Schema(
    {
        Required("package_version"): PACKAGE_VERSION_SCHEMA,
        Optional("state"): Schema(
            {
                Optional("resolved_dependencies"): All([PACKAGE_VERSION_REQUIRED_NAME_SCHEMA], Length(min=1)),
                Optional("package_version_from"): All([PACKAGE_VERSION_REQUIRED_NAME_SCHEMA], Length(min=1)),
                Optional("package_version_from_allow_other"): bool,
            }
        ),
    }
)

PRESCRIPTION_STEP_RUN_SCHEMA = Schema(
    {
        Optional("score"): Optional(float),
        Optional("justification"): All([JUSTIFICATION_SCHEMA], Length(min=1)),
        Optional("not_acceptable"): _NONEMPTY_STRING,
        Optional("eager_stop_pipeline"): _NONEMPTY_STRING,
        Optional("multi_package_resolution"): bool,
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)

PRESCRIPTION_STEP_SCHEMA = Schema(
    {
        Required("match"): Any(
            All([PRESCRIPTION_STEP_MATCH_ENTRY_SCHEMA], Length(min=1)), PRESCRIPTION_STEP_MATCH_ENTRY_SCHEMA
        ),
        Required("run"): PRESCRIPTION_STEP_RUN_SCHEMA,
        Required("type"): "step",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)


#
# Skip package step unit.
#

PRESCRIPTION_SKIP_PACKAGE_STEP_MATCH_ENTRY_SCHEMA = PRESCRIPTION_STEP_MATCH_ENTRY_SCHEMA


PRESCRIPTION_SKIP_PACKAGE_STEP_RUN_SCHEMA = Schema(
    {
        Optional("multi_package_resolution"): bool,
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)


PRESCRIPTION_SKIP_PACKAGE_STEP_SCHEMA = Schema(
    {
        Required("match"): Any(
            PRESCRIPTION_SKIP_PACKAGE_STEP_MATCH_ENTRY_SCHEMA,
        ),
        Optional("run"): PRESCRIPTION_SKIP_PACKAGE_STEP_RUN_SCHEMA,
        Required("type"): "step.SkipPackage",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)


#
# Add package step unit.
#

PRESCRIPTION_ADD_PACKAGE_STEP_MATCH_ENTRY_SCHEMA = PRESCRIPTION_STEP_MATCH_ENTRY_SCHEMA


PRESCRIPTION_ADD_PACKAGE_STEP_RUN_SCHEMA = Schema(
    {
        Optional("multi_package_resolution"): bool,
        Required("package_version"): Schema(
            {
                Required("name"): _python_package_name,
                Required("locked_version"): _locked_version,
                Required("index_url"): _NONEMPTY_STRING,
                Required("develop"): bool,
            }
        ),
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)


PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA = Schema(
    {
        Required("match"): Any(
            All([PRESCRIPTION_ADD_PACKAGE_STEP_MATCH_ENTRY_SCHEMA], Length(min=1)),
            PRESCRIPTION_ADD_PACKAGE_STEP_MATCH_ENTRY_SCHEMA,
        ),
        Required("run"): PRESCRIPTION_ADD_PACKAGE_STEP_RUN_SCHEMA,
        Required("type"): "step.AddPackage",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)


#
# Group step unit.
#

PRESCRIPTION_GROUP_STEP_RUN_SCHEMA = Schema(
    {
        Optional("score"): Optional(float),
        Optional("justification"): All([JUSTIFICATION_SCHEMA], Length(min=1)),
        Optional("not_acceptable"): _NONEMPTY_STRING,
        Optional("eager_stop_pipeline"): _NONEMPTY_STRING,
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)


PRESCRIPTION_GROUP_STEP_MATCH_SCHEMA = Schema(
    {Required("group"): All([{Required("package_version"): PACKAGE_VERSION_REQUIRED_NAME_SCHEMA}], Length(min=2))}
)


PRESCRIPTION_GROUP_STEP_SCHEMA = Schema(
    {
        Required("match"): Any(
            All([PRESCRIPTION_GROUP_STEP_MATCH_SCHEMA], Length(min=1)),
            PRESCRIPTION_GROUP_STEP_MATCH_SCHEMA,
        ),
        Required("run"): PRESCRIPTION_GROUP_STEP_RUN_SCHEMA,
        Required("type"): "step.Group",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)


#
# Stride unit.
#

PRESCRIPTION_STRIDE_MATCH_ENTRY_SCHEMA = Schema(
    {Required("state"): Schema({Required("resolved_dependencies"): [PACKAGE_VERSION_SCHEMA]})}
)

PRESCRIPTION_STRIDE_RUN_SCHEMA = Schema(
    {
        Optional("not_acceptable"): _NONEMPTY_STRING,
        Optional("eager_stop_pipeline"): _NONEMPTY_STRING,
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)

PRESCRIPTION_STRIDE_SCHEMA = Schema(
    {
        Required("match"): Any(
            All([PRESCRIPTION_STRIDE_MATCH_ENTRY_SCHEMA], Length(min=1)), PRESCRIPTION_STRIDE_MATCH_ENTRY_SCHEMA
        ),
        Required("run"): PRESCRIPTION_STRIDE_RUN_SCHEMA,
        Required("type"): "stride",
        **_UNIT_SCHEMA_BASE_DICT,
    }
)

#
# Wrap unit.
#

PRESCRIPTION_WRAP_MATCH_ENTRY_SCHEMA = Schema(
    {Optional("state"): Schema({Required("resolved_dependencies"): [PACKAGE_VERSION_SCHEMA]})}
)

PRESCRIPTION_WRAP_RUN_SCHEMA = Schema(
    {
        Optional("justification"): All([JUSTIFICATION_SCHEMA], Length(min=1)),
        Optional("advised_manifest_changes"): object,
        **_UNIT_RUN_SCHEMA_BASE_DICT,
    }
)

PRESCRIPTION_WRAP_SCHEMA = Schema(
    {
        Required("run"): PRESCRIPTION_WRAP_RUN_SCHEMA,
        Required("type"): "wrap",
        Optional("match"): Any(
            All([PRESCRIPTION_WRAP_MATCH_ENTRY_SCHEMA], Length(min=1)), PRESCRIPTION_WRAP_MATCH_ENTRY_SCHEMA
        ),
        **_UNIT_SCHEMA_BASE_DICT,
    }
)


#
# GitHub release notes wrap unit.
#

PRESCRIPTION_GH_RELEASE_NOTES_WRAP_MATCH_ENTRY_SCHEMA = Schema(
    {
        Required("state"): Schema(
            {
                Required("resolved_dependencies"): Any(
                    All([PACKAGE_VERSION_REQUIRED_NAME_SCHEMA], Length(min=1, max=1)),
                    PACKAGE_VERSION_REQUIRED_NAME_SCHEMA,
                )
            }
        )
    }
)

PRESCRIPTION_GH_RELEASE_NOTES_WRAP_RUN_ENTRY_SCHEMA = Schema(
    {
        Required("organization"): _NONEMPTY_STRING,
        Optional("tag_version_prefix"): _NONEMPTY_STRING,
        Required("repository"): _NONEMPTY_STRING,
    }
)


PRESCRIPTION_GH_RELEASE_NOTES_WRAP_RUN_ENTRIES_SCHEMA = Schema(
    {"release_notes": PRESCRIPTION_GH_RELEASE_NOTES_WRAP_RUN_ENTRY_SCHEMA}
)


PRESCRIPTION_GH_RELEASE_NOTES_WRAP_SCHEMA = Schema(
    {
        Required("type"): "wrap.GHReleaseNotes",
        Required("match"): Any(
            All([PRESCRIPTION_GH_RELEASE_NOTES_WRAP_MATCH_ENTRY_SCHEMA], Length(min=1)),
            PRESCRIPTION_GH_RELEASE_NOTES_WRAP_MATCH_ENTRY_SCHEMA,
        ),
        Required("run"): PRESCRIPTION_GH_RELEASE_NOTES_WRAP_RUN_ENTRIES_SCHEMA,
        **_UNIT_SCHEMA_BASE_DICT,
    }
)

#
# Prescription.
#

PRESCRIPTION_UNITS_SCHEMA = Schema(
    {
        Optional("boots"): [PRESCRIPTION_BOOT_SCHEMA],
        Optional("sieves"): [Any(PRESCRIPTION_SIEVE_SCHEMA, PRESCRIPTION_SKIP_PACKAGE_SIEVE_SCHEMA)],
        Optional("steps"): [
            Any(
                PRESCRIPTION_STEP_SCHEMA,
                PRESCRIPTION_SKIP_PACKAGE_STEP_SCHEMA,
                PRESCRIPTION_ADD_PACKAGE_STEP_SCHEMA,
                PRESCRIPTION_GROUP_STEP_SCHEMA,
            )
        ],
        Optional("pseudonyms"): [PRESCRIPTION_PSEUDONYM_SCHEMA],
        Optional("strides"): [PRESCRIPTION_STRIDE_SCHEMA],
        Optional("wraps"): [
            Any(
                PRESCRIPTION_WRAP_SCHEMA,
                PRESCRIPTION_GH_RELEASE_NOTES_WRAP_SCHEMA,
            )
        ],
    }
)

PRESCRIPTION_SCHEMA = Schema(
    {
        Required("units"): PRESCRIPTION_UNITS_SCHEMA,
        Optional("metadata"): _METADATA_SCHEMA,
    }
)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Sieves that filters packages available on Operate First Pulp instance."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

from thoth.python import PackageVersion
from .index_label import PulpIndexLabelSieve

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


class NoPulpIndexLabelSieve(PulpIndexLabelSieve):
    """A sieve that filters out packages that come from Operate First Pulp instance."""

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[Any, Any], None, None]:
        """Include this wrap in adviser, if label pulp-index=disabled is used."""
        if not builder_context.is_included(cls) and builder_context.labels.get("opf-pulp-indexes") == "disabled":
            yield {}
            return None

        yield from ()
        return None

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Remove any packages that do not come from the Pulp instance."""
        from . import PULP_URL

        if not self._stack_info_reported:
            self.context.stack_info.append(
                {
                    "type": "INFO",
                    "message": "Ignoring packages provided by the Operate First Pulp instance",
                    "link": PULP_URL,
                }
            )

        for package_version in package_versions:
            if not package_version.index.url.startswith(PULP_URL):
                yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of generic v1 prescription unit handling."""

import logging
from typing import Type

import pytest

from thoth.adviser.exceptions import NotAcceptable
from thoth.adviser.exceptions import EagerStopPipeline
from thoth.adviser.context import Context
from thoth.adviser.prescription.v1 import UnitPrescription

from ...base import AdviserTestCase


class AdviserUnitPrescriptionTestCase(AdviserTestCase):
    """Test implementation of generic v1 prescription unit handling."""

    @staticmethod
    def setup_method() -> None:
        """Make sure prescription is not assigned before any test."""
        UnitPrescription._PRESCRIPTION = None

    @staticmethod
    def teardown_method() -> None:
        """Make sure prescription is not assigned before any test."""
        UnitPrescription._PRESCRIPTION = None

    def check_run_stack_info(self, context, unit_class: Type[UnitPrescription], **kwargs) -> None:
        """Check assigning stack info."""
        assert not context.stack_info

        unit = unit_class()
        unit.pre_run()
        with unit.assigned_context(context):
            result = unit.run(**kwargs)
            if result is not None:
                list(result)

        assert self.verify_justification_schema(context.stack_info)
        assert context.stack_info == unit.run_prescription["stack_info"]

    @staticmethod
    def check_run_eager_stop_pipeline(context: Context, unit_class: Type[UnitPrescription], **kwargs) -> None:
        """Check eager stop pipeline configuration."""
        unit = unit_class()
        unit.pre_run()
        with unit.assigned_context(context):
            with pytest.raises(EagerStopPipeline, match=unit.run_prescription["eager_stop_pipeline"]):
                result = unit.run(**kwargs)
                if result is not None:
                    list(result)

    @staticmethod
    def check_run_not_acceptable(context: Context, unit_class: Type[UnitPrescription], **kwargs) -> None:
        """Check raising not acceptable."""
        unit = unit_class()
        unit.pre_run()
        with unit.assigned_context(context):
            with pytest.raises(NotAcceptable, match=unit.run_prescription["not_acceptable"]):
                result = unit.run(**kwargs)
                if result is not None:
                    list(result)

    @staticmethod
    def check_run_log(caplog, context: Context, log_level: str, unit_class: Type[UnitPrescription], **kwargs) -> None:
        """Check raising not acceptable."""
        unit = unit_class()
        unit.pre_run()
        with unit.assigned_context(context):
            with caplog.at_level(getattr(logging, log_level)):
                result = unit.run(**kwargs)
                if result is not None:
                    list(result)

        assert f"{unit.get_unit_name()}: {unit.run_prescription['log']['message']}" in caplog.messages



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of Monte-Carlo Tree Search (MCTS) based predictor with annealing schedule."""

import math

from flexmock import flexmock
import pytest

from thoth.adviser.context import Context
from thoth.adviser.predictors import MCTS
from thoth.adviser.predictors import TemporalDifference

from ..base import AdviserTestCase


class TestMCTS(AdviserTestCase):
    """Test implementation of MCTS based predictor with annealing schedule."""

    def test_init(self) -> None:
        """Test the initialization part."""
        predictor = MCTS()
        assert predictor._policy == {}
        assert predictor._temperature_history == []
        assert predictor._temperature == 0.0
        assert predictor._next_state is None

    def test_pre_run(self) -> None:
        """Test calling pre-run for the initialization part."""
        flexmock(TemporalDifference)
        TemporalDifference.should_receive("pre_run").once()

        state = flexmock()
        predictor = MCTS()
        assert predictor._next_state is None
        predictor._next_state = state
        predictor.pre_run()
        assert predictor._next_state is None

    def test_set_reward_signal_nan(self) -> None:
        """Test the reward signal if no new state was generated."""
        predictor = MCTS()

        state = flexmock()

        assert predictor._policy == {}
        predictor._next_state = flexmock()
        predictor.set_reward_signal(state, ("numpy", "2.0.0", "https://pypi.org/simple"), math.nan)
        assert predictor._next_state is None
        assert predictor._policy == {}

    def test_set_reward_signal(self) -> None:
        """Test the reward signal if no new state was generated."""
        predictor = MCTS()

        state = flexmock()

        assert predictor._policy == {}
        predictor._next_state = flexmock()

        predictor.set_reward_signal(state, ("numpy", "2.0.0", "https://pypi.org/simple"), 7355608.0)
        assert predictor._next_state is state
        assert predictor._policy == {}

    def test_set_reward_signal_inf(self) -> None:
        """Test the reward signal if a final state was generated."""
        predictor = MCTS()

        state = flexmock(score=3.1)
        state.should_receive("iter_resolved_dependencies").and_return(
            [
                ("numpy", "2.0.0", "https://pypi.org/simple"),
                ("tensorflow", "2.0.0", "https://thoth-station.ninja/simple"),
            ]
        )
        # numpy was already seen, tensorflow was not seen yet
        predictor._policy = {
            ("numpy", "2.0.0", "https://pypi.org/simple"): [2.3, 100],
        }
        predictor._next_state = flexmock()
        predictor.set_reward_signal(state, ("numpy", "2.0.0", "https://pypi.org/simple"), math.inf)
        assert predictor._next_state is None
        assert predictor._policy == {
            ("numpy", "2.0.0", "https://pypi.org/simple"): [2.3 + 3.1, 101],
            ("tensorflow", "2.0.0", "https://thoth-station.ninja/simple"): [3.1, 1],
        }

    @pytest.mark.parametrize("next_state", [None, flexmock()])
    def test_run_heat_up(self, context: Context, next_state) -> None:
        """Test running the predictor in the "heat-up" phase regardless next state being set."""
        state = flexmock()
        unresolved_dependency = ("tensorflow", "2.0.0", "https://pypi.org/simple")

        predictor = MCTS()
        predictor._next_state = None

        flexmock(TemporalDifference)
        TemporalDifference.should_receive("run").with_args().and_return(state, unresolved_dependency).once()

        context.iteration = 1  # Some small number to hit the heat-up part.
        with predictor.assigned_context(context):
            assert predictor.run() == (state, unresolved_dependency)

    def test_run_next_state(self, context: Context) -> None:
        """Test running the predictor when the next state is scheduled."""
        state = flexmock()
        unresolved_dependency = ("tensorflow", "2.0.0", "https://pypi.org/simple")
        state.should_receive("get_random_unresolved_dependency").with_args(prefer_recent=True).and_return(
            unresolved_dependency
        ).once()

        predictor = MCTS()
        predictor._next_state = state
        context.beam.should_receive("get_last").and_return(state).once()
        context.iteration = 1000000  # Some big number not to hit the heat-up part.
        with predictor.assigned_context(context):
            assert predictor.run() == (state, unresolved_dependency)

    def test_run_next_state_no_last(self, context: Context) -> None:
        """Test running the predictor when the next state is not last state added to beam."""
        state = flexmock()
        unresolved_dependency = ("tensorflow", "2.0.0", "https://pypi.org/simple")

        predictor = MCTS()
        predictor._next_state = flexmock()
        context.beam.should_receive("get_last").and_return(flexmock()).once()

        flexmock(TemporalDifference)
        TemporalDifference.should_receive("run").with_args().and_return(state, unresolved_dependency).once()

        context.iteration = 1000000  # Some big number not to hit the heat-up part.
        with predictor.assigned_context(context):
            assert predictor.run() == (state, unresolved_dependency)

    def test_run_no_next_state(self, context: Context) -> None:
        """Test running the predictor when no next state is scheduled."""
        predictor = MCTS()
        assert predictor._next_state is None

        # If no next state kept, we follow logic from the TD-learning.
        flexmock(TemporalDifference)
        state = flexmock()
        unresolved_dependency = flexmock()
        TemporalDifference.should_receive("run").with_args().and_return(state, unresolved_dependency).once()
        context.iteration = 1000000  # Some big number not to hit the heat-up part.
        with predictor.assigned_context(context):
            assert predictor.run() == (state, unresolved_dependency)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test removing pinned versions in direct dependencies."""

from flexmock import flexmock

from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import CutLockedSieve
from thoth.python import Source
from thoth.python import PackageVersion
from thoth.python import Project

from ..base import AdviserUnitTestCase


class TestCutPreReleasesSieve(AdviserUnitTestCase):
    """Test removing pinned versions in direct dependencies."""

    _CASE_PIPFILE_LOCKED = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "==1.9.0"

[dev-packages]
pytest = "==5.3.1"
"""

    _CASE_PIPFILE_NOT_LOCKED = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[dev-packages]
pytest = "*"

[pipenv]
allow_prereleases = true
"""

    UNIT_TESTED = CutLockedSieve

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_cut_locked(self) -> None:
        """Test removing a locked package based on direct dependencies."""
        tf = PackageVersion(
            name="tensorflow",
            version="==2.0.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [tf])) == []

    def test_cut_locked_dev(self) -> None:
        """Test removing a locked package based on direct dev dependencies."""
        pytest = PackageVersion(
            name="pytest",
            version="==2.0.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [pytest])) == []

    def test_no_cut(self) -> None:
        """Test not removing a locked package based on direct dependencies."""
        tf = PackageVersion(
            name="tensorflow",
            version="==1.9.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [tf])) == [tf]

    def test_no_cut_dev(self) -> None:
        """Test not removing a locked package based on dev direct dependencies."""
        pytest = PackageVersion(
            name="pytest",
            version="==5.3.1",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [pytest])) == [pytest]

    def test_noop(self) -> None:
        """Test no operation if dependencies are not locked."""
        tf = PackageVersion(
            name="tensorflow",
            version="==1.9.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_NOT_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [tf])) == [tf]

    def test_noop_dev(self) -> None:
        """Test no operation if dependencies are not locked."""
        pytest = PackageVersion(
            name="pytest",
            version="==5.3.1",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_NOT_LOCKED))
        with CutLockedSieve.assigned_context(context):
            sieve = CutLockedSieve()
            assert list(sieve.run(p for p in [pytest])) == [pytest]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A prescription sieve implementing skipping a package in a dependency graph."""

from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING
import logging

import attr
from thoth.python import PackageVersion
from voluptuous import Any as SchemaAny
from voluptuous import Schema
from voluptuous import Required

from .unit import UnitPrescription
from .schema import PRESCRIPTION_SKIP_PACKAGE_SIEVE_RUN_ENTRY_SCHEMA
from .schema import PRESCRIPTION_SKIP_PACKAGE_SIEVE_MATCH_ENTRY_SCHEMA

from ...exceptions import SkipPackage

if TYPE_CHECKING:
    from ...pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SkipPackageSievePrescription(UnitPrescription):
    """Skip package sieve prescription unit implementation."""

    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("package_name"): SchemaAny(str, None),
            Required("run"): SchemaAny(PRESCRIPTION_SKIP_PACKAGE_SIEVE_RUN_ENTRY_SCHEMA, None),
            Required("match"): PRESCRIPTION_SKIP_PACKAGE_SIEVE_MATCH_ENTRY_SCHEMA,
            Required("prescription"): Schema({"run": bool}),
        }
    )

    @staticmethod
    def is_sieve_unit_type() -> bool:
        """Check if this unit is of type sieve."""
        return True

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Check if the given pipeline unit should be included in the given pipeline configuration."""
        if cls._should_include_base(builder_context):
            prescription: Dict[str, Any] = cls._PRESCRIPTION  # type: ignore
            prescription_conf = {"run": False}
            if isinstance(prescription["match"], list):
                for item in prescription["match"]:
                    yield {
                        "package_name": item["package_name"],
                        "run": prescription.get("run"),
                        "match": item,
                        "prescription": prescription_conf,
                    }
            else:
                yield {
                    "package_name": prescription["match"]["package_name"],
                    "run": prescription.get("run"),
                    "match": prescription["match"],
                    "prescription": prescription_conf,
                }
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this unit before each run."""
        super().pre_run()

    def run(self, _: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Run main entry-point for sieves to filter and score packages."""
        prescription_conf = self._configuration["prescription"]
        if not prescription_conf["run"]:
            self._run_log()
            self._run_stack_info()
            prescription_conf["run"] = True

        raise SkipPackage



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A step that is setting score for packages."""

from typing import Any
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import List
from typing import Dict
from typing import TYPE_CHECKING
import logging
import random

import attr
from thoth.python import PackageVersion
from voluptuous import Any as SchemaAny
from voluptuous import Optional as SchemaOptional
from voluptuous import Required
from voluptuous import Schema

from ...state import State
from ...step import Step


if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SetScoreStep(Step):
    """A step that is setting score for packages."""

    # Assign probability is used to "assign" a score to the package to simulate knowledge
    # coverage for packages resolved - 0.75 means ~75% of packages will have a score.
    CONFIGURATION_SCHEMA: Schema = Schema(
        {
            Required("multi_package_resolution"): bool,
            Required("package_name"): str,
            SchemaOptional("index_url"): SchemaAny(str, None),
            SchemaOptional("package_version"): SchemaAny(str, None),
            SchemaOptional("score"): SchemaAny(float, None),
        }
    )
    CONFIGURATION_DEFAULT: Dict[str, Any] = {
        "index_url": None,
        "multi_package_resolution": False,
        "package_name": None,
        "package_version": None,
        "score": None,
    }

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, never."""
        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this pipeline unit before each run."""
        if self.configuration["score"] is None:
            self.configuration["score"] = random.uniform(self.SCORE_MIN, self.SCORE_MAX)

        super().pre_run()

    def run(
        self, _: State, package_version: PackageVersion
    ) -> Optional[Tuple[Optional[float], Optional[List[Dict[str, str]]]]]:
        """Score the given package."""
        if (
            self.configuration["package_version"] is not None
            and package_version.locked_version != self.configuration["package_version"]
        ) or (
            self.configuration["index_url"] is not None and package_version.index.url != self.configuration["index_url"]
        ):
            return None

        return self.configuration["score"], None



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A sieve to filter out packages coming from other indexes than the ones configured."""

import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.python import PackageVersion
from voluptuous import Schema
from voluptuous import Required

from ..sieve import Sieve

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class FilterConfiguredIndexSieve(Sieve):
    """A sieve to filter out packages coming from other indexes than the ones configured."""

    CONFIGURATION_DEFAULT = {"package_name": None, "allowed_indexes": None}
    CONFIGURATION_SCHEMA: Schema = Schema({Required("package_name"): None, Required("allowed_indexes"): {str}})
    _JUSTIFICATION_LINK = jl("index_conf")

    packages_seen = attr.ib(type=Set[Tuple[str, str, str]], default=attr.Factory(set), init=False)

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Enable this pipeline unit if the adjustment is enabled."""
        if not builder_context.is_included(cls) and builder_context.project.pipfile.thoth.disable_index_adjustment:
            allowed_indexes = set()
            for source in builder_context.project.pipfile.meta.sources.values():
                allowed_indexes.add(source.url)

            yield {
                "package_name": None,
                "allowed_indexes": allowed_indexes,
            }
            return None

        yield from ()
        return None

    def pre_run(self) -> None:
        """Initialize this unit before each run."""
        self.packages_seen.clear()
        super().pre_run()

    def to_dict(self) -> Dict[str, Any]:
        """Turn this pipeline step into its dictionary representation, override to have unit JSON serializable."""
        configuration = {}

        for key, value in self.configuration.items():
            configuration[key] = value if not isinstance(value, set) else sorted(value)

        return {"name": self.__class__.__name__, "configuration": configuration, "unit_run": self.unit_run}

    def run(self, package_versions: Generator[PackageVersion, None, None]) -> Generator[PackageVersion, None, None]:
        """Cut-off packages that are not coming from an allowed index."""
        for package_version in package_versions:
            if package_version.index.url not in self.configuration["allowed_indexes"]:
                package_tuple = package_version.to_tuple()

                if package_tuple not in self.packages_seen:
                    self.packages_seen.add(package_tuple)
                    msg = f"Removing package {package_tuple} as it does not use enabled index"
                    _LOGGER.warning("%s - see %s", msg, self._JUSTIFICATION_LINK)
                    self.context.stack_info.append(
                        {
                            "type": "WARNING",
                            "message": msg,
                            "link": self._JUSTIFICATION_LINK,
                        }
                    )

                continue

            yield package_version



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to check for solved software environment before running any resolution."""

import logging
import os
from typing import Generator
from typing import Dict
from typing import Any
from typing import TYPE_CHECKING

import attr
from thoth.common import get_justification_link as jl
from thoth.common import OpenShift

from ..boot import Boot
from ..exceptions import NotAcceptable

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SolvedSoftwareEnvironmentBoot(Boot):
    """A boot to check for solved software environment before running any resolution."""

    _JUSTIFICATION_LINK = jl("solved_sw_env")
    _THOTH_ADVISER_DEPLOYMENT_CONFIGURED_SOLVERS = os.getenv("THOTH_ADVISER_DEPLOYMENT_CONFIGURED_SOLVERS", "")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self, always."""
        if builder_context.project.runtime_environment.is_fully_specified() and not builder_context.is_included(cls):
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Check for version clash in packages."""
        if self.context.graph.solved_software_environment_exists(
            os_name=self.context.project.runtime_environment.operating_system.name,
            os_version=self.context.project.runtime_environment.operating_system.version,
            python_version=self.context.project.runtime_environment.python_version,
        ):
            return

        runtime_environment = self.context.project.runtime_environment
        msg = (
            f"No observations found for {runtime_environment.operating_system.name!r} in "
            f"version {runtime_environment.operating_system.version!r} using "
            f"Python {runtime_environment.python_version!r}"
        )

        self.context.stack_info.append(
            {
                "type": "ERROR",
                "message": msg,
                "link": self._JUSTIFICATION_LINK,
            }
        )

        _LOGGER.warning("%s - %s", msg, self._JUSTIFICATION_LINK)
        _LOGGER.warning("Available configurations:")

        configurations = []
        solvers = self._THOTH_ADVISER_DEPLOYMENT_CONFIGURED_SOLVERS.split()
        for solver in solvers:
            solver = solver.strip()
            if not solver:
                continue

            item = OpenShift.parse_python_solver_name(solver)
            configurations.append(item)

            if item["os_name"] == "rhel":
                # Duplicate entry as we can also guide on the same UBI environment. UBI and RHEL are binary compatible.
                other_item = dict(item)
                other_item["os_name"] = "ubi"
                configurations.append(other_item)

        _LOGGER.warning("{:<16} {:<16} {:<8}".format("OS name", "OS version", "Python version"))
        for conf in sorted(
            configurations,
            key=lambda i: (i["os_name"], i["os_version"], i["python_version"]),
        ):
            self.context.stack_info.append(
                {
                    "message": f"Consider using {conf['os_name']!r} in version {conf['os_version']!r} "
                    f"with Python {conf['python_version']}",
                    "type": "ERROR",
                    "link": self._JUSTIFICATION_LINK,
                }
            )
            _LOGGER.warning("{:<16} {:<16} {:<8}".format(conf["os_name"], conf["os_version"], conf["python_version"]))

        raise NotAcceptable(msg)



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test boot for Python version assignment."""

from flexmock import flexmock

from thoth.adviser.boots import PythonVersionBoot
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.common import get_justification_link as jl
from thoth.python import Project

from ..base import AdviserUnitTestCase


class TestPythonVersionBoot(AdviserUnitTestCase):
    """Test changes to runtime environment or configuration done with respect to configured Python."""

    _CASE_PIPFILE_NO_PYTHON = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"
"""

    _CASE_PIPFILE_PYTHON = """
[[source]]
url = "https://pypi.org/simple"
verify_ssl = true
name = "pypi"

[packages]
tensorflow = "*"

[requires]
python_version = "3.6"
"""

    UNIT_TESTED = PythonVersionBoot

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_no_python_config(self) -> None:
        """Test assigning Python version from Thoth's config file."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_PYTHON), stack_info=[])
        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8"

        assert context.project.runtime_environment.python_version is None

        boot = PythonVersionBoot()
        with PythonVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "rhel"
        assert context.project.runtime_environment.operating_system.version == "8"
        assert context.project.runtime_environment.python_version == "3.6"
        assert context.stack_info == [
            {
                "message": "No version of Python specified in the configuration, using "
                "Python version found in Pipfile: '3.6'",
                "type": "WARNING",
                "link": jl("py_version"),
            },
        ]

    def test_no_python_pipfile(self) -> None:
        """Test assigning Python version from Pipfile."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_NO_PYTHON), stack_info=[])
        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8"
        context.project.runtime_environment.python_version = "3.6"

        boot = PythonVersionBoot()
        with PythonVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "rhel"
        assert context.project.runtime_environment.operating_system.version == "8"
        assert context.project.runtime_environment.python_version == "3.6"
        assert context.stack_info == [
            {
                "message": "No version of Python specified explicitly, assigning the one "
                "found in Thoth's configuration: '3.6'",
                "type": "WARNING",
                "link": jl("py_version"),
            }
        ]

    def test_python_version_mismatch(self) -> None:
        """Test when python version stated in Pipfile does not match with the one provided in the configuration."""
        context = flexmock(project=Project.from_strings(self._CASE_PIPFILE_PYTHON), stack_info=[])
        context.project.runtime_environment.operating_system.name = "rhel"
        context.project.runtime_environment.operating_system.version = "8"
        context.project.runtime_environment.python_version = "3.8"

        boot = PythonVersionBoot()
        with PythonVersionBoot.assigned_context(context):
            boot.run()

        assert context.project.runtime_environment.operating_system.name == "rhel"
        assert context.project.runtime_environment.operating_system.version == "8"
        assert context.project.runtime_environment.python_version == "3.8"
        assert context.stack_info == [
            {
                "message": "Python version stated in Pipfile ('3.6') does not match with the one "
                "specified in the Thoth configuration ('3.8'), using Python version from Thoth "
                "configuration implicitly",
                "type": "WARNING",
                "link": jl("py_version"),
            }
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""A boot to notify about runtime environments not supported by solvers enabled in a deployment."""

import os
import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import TYPE_CHECKING

import attr
from thoth.common.exceptions import SolverNameParseError
from thoth.common import get_justification_link as jl
from thoth.common import normalize_os_version
from thoth.common import OpenShift

from ..boot import Boot

if TYPE_CHECKING:
    from ..pipeline_builder import PipelineBuilderContext

_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class SolversConfiguredBoot(Boot):
    """A boot to notify about runtime environments not supported by solvers enabled in a deployment."""

    _SOLVERS_CONFIGURED = os.getenv("THOTH_ADVISER_DEPLOYMENT_CONFIGURED_SOLVERS")
    _JUSTIFICATION = jl("eol_env")

    @classmethod
    def should_include(cls, builder_context: "PipelineBuilderContext") -> Generator[Dict[str, Any], None, None]:
        """Register self if users use runtime environments not matching solvers in the configmap."""
        if not cls._SOLVERS_CONFIGURED or builder_context.is_included(cls):
            yield from ()
            return None

        runtime_environment = (
            builder_context.project.runtime_environment.operating_system.name,
            normalize_os_version(
                builder_context.project.runtime_environment.operating_system.name,
                builder_context.project.runtime_environment.operating_system.version,
            ),
            builder_context.project.python_version,
        )
        for solver_name in cls._SOLVERS_CONFIGURED.splitlines():
            solver_name = solver_name.strip()

            if not solver_name:
                continue

            try:
                solver_info = OpenShift.parse_python_solver_name(solver_name)
            except SolverNameParseError:
                _LOGGER.exception("Please report this error to Thoth service administrator")
                yield from ()
                return None

            if runtime_environment == (
                solver_info["os_name"],
                solver_info["os_version"],
                solver_info["python_version"],
            ):
                break
        else:
            yield {}
            return None

        yield from ()
        return None

    def run(self) -> None:
        """Inform about runtime environment that is not backed up by a solver."""
        self.context.stack_info.append(
            {
                "type": "WARNING",
                "message": "Runtime environment used is no longer supported, it is "
                "recommended to switch to another runtime environment",
                "link": self._JUSTIFICATION,
            }
        )



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test version checking boot."""

from thoth.adviser.boots import VersionCheckBoot
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.enums import RecommendationType
from thoth.common import get_justification_link as jl

from ..base import AdviserUnitTestCase


class TestVersionCheckBoot(AdviserUnitTestCase):
    """A boot that checks if versions are too lax."""

    UNIT_TESTED = VersionCheckBoot

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(recommendation_type=RecommendationType.LATEST)
        self.verify_multiple_should_include(builder_context)

    def test_no_version(self, context: Context) -> None:
        """Test providing stack information if there is no version identifier."""
        context.project.pipfile.packages.packages.clear()
        context.project.pipfile.dev_packages.packages.clear()

        context.project.add_package("flask")
        assert not context.stack_info

        boot = self.UNIT_TESTED()
        with boot.assigned_context(context):
            boot.run()

        assert context.stack_info == [
            {
                "link": jl("lax_version"),
                "message": "No version range specifier for 'flask' found, it is recommended to "
                "specify version ranges in requirements",
                "type": "WARNING",
            },
        ]

    def test_lax_version(self, context: Context) -> None:
        """Test providing stack information if the used version is too lax."""
        context.project.pipfile.packages.packages.clear()
        context.project.pipfile.dev_packages.packages.clear()

        context.project.add_package("flask", ">=0.12")
        assert not context.stack_info

        boot = self.UNIT_TESTED()
        with boot.assigned_context(context):
            boot.run()

        assert context.stack_info == [
            {
                "link": jl("lax_version"),
                "message": "Version range specifier ('>=0.12') for 'flask' might be too lax",
                "type": "WARNING",
            },
        ]

    def test_version_ok(self, context: Context) -> None:
        """Test not providing stack information if the used version is okay."""
        context.project.pipfile.packages.packages.clear()
        context.project.pipfile.dev_packages.packages.clear()

        context.project.add_package("flask", ">=0.12,<2.0.0")
        assert not context.stack_info

        boot = self.UNIT_TESTED()
        with boot.assigned_context(context):
            boot.run()

        assert context.stack_info == []



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test implementation of prescription handling."""

import pytest
import yaml

from thoth.adviser.exceptions import PrescriptionSchemaError
from thoth.adviser.exceptions import PrescriptionDuplicateUnitNameError
from thoth.adviser.prescription import Prescription

from ...base import AdviserTestCase


class TestPrescription(AdviserTestCase):
    """Test implementation of prescription handling."""

    def test_load(self) -> None:
        """Test loading prescription."""
        prescription_path = str(self.data_dir / "prescriptions")
        instance = Prescription.load(prescription_path)
        assert instance is not None
        assert list(instance.boots_dict) == ["thoth.BootUnit"]
        assert list(instance.pseudonyms_dict) == ["thoth.PseudonymUnit"]
        assert list(instance.sieves_dict) == ["thoth.SieveUnit"]
        assert list(instance.steps_dict) == ["thoth.StepUnit"]
        assert list(instance.strides_dict) == ["thoth.StrideUnit"]
        assert list(instance.wraps_dict) == ["thoth.WrapUnit"]

        assert [u.get_unit_name() for u in instance.iter_boot_units()] == ["thoth.BootUnit"]
        assert [u.get_unit_name() for u in instance.iter_pseudonym_units()] == ["thoth.PseudonymUnit"]
        assert [u.get_unit_name() for u in instance.iter_sieve_units()] == ["thoth.SieveUnit"]
        assert [u.get_unit_name() for u in instance.iter_step_units()] == ["thoth.StepUnit"]
        assert [u.get_unit_name() for u in instance.iter_stride_units()] == ["thoth.StrideUnit"]
        assert [u.get_unit_name() for u in instance.iter_wrap_units()] == ["thoth.WrapUnit"]

    def test_from_dict_validate_error(self) -> None:
        """Test raising an error if schema validation fails."""
        with pytest.raises(PrescriptionSchemaError):
            Prescription.from_dict({"foo": "bar"}, prescription_name="thoth", prescription_release="2021.06.15")

    def test_from_dict_duplicate_name_error(self) -> None:
        """Test raising duplicate error if unit names clash."""
        unit = {
            "name": "BootUnit",
            "type": "boot",
            "should_include": {"adviser_pipeline": True},
            "run": {"log": {"message": "Some text printed", "type": "INFO"}},
        }

        prescription = {
            "units": {
                "boots": [unit, dict(unit)],
            },
        }

        with pytest.raises(PrescriptionDuplicateUnitNameError):
            Prescription.from_dict(prescription, prescription_name="thoth", prescription_release="2021.06.15")

    def test_yaml_cloader(self) -> None:
        """Test loading using yaml.CLoader."""
        content = yaml.load("foo: bar", Loader=yaml.CLoader)
        assert content == {"foo": "bar"}



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Utilities which wrap adviser's and dependency monkey's and run them in a subprocess.

This is useful for a cluster deployment where we restrict resources allocated for an adviser or dependency monkey run.
Annealing is run as a subprocess to the main process - if resources are exhausted (CPU time or memory allocated), the
subprocess is killed (either by liveness probe or by OOM killer) and the main process can still produce an
error message.
"""

from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Union
from typing import Optional
import logging
import os
import time

from thoth.common import get_justification_link as jl

from .exceptions import CannotProduceStack
from .exceptions import UnresolvedDependencies
from .dependency_monkey import DependencyMonkey
from .dm_report import DependencyMonkeyReport
from .report import Report
from .resolver import Resolver

_LOGGER = logging.getLogger(__name__)
# This file is created by OpenShift's liveness probe on timeout.
_LIVENESS_PROBE_KILL_FILE = "/tmp/thoth_adviser_cpu_timeout"
_FORK = bool(int(os.getenv("THOTH_ADVISER_FORK", 0)))


def subprocess_run(
    resolver: Union[Resolver, DependencyMonkey],
    print_func: Callable[[float, Union[Dict[str, Any], List[Any]]], None],
    result_dict: Dict[str, Any],
    plot: Optional[str] = None,
    *,
    with_devel: bool = True,
    verbose: bool = False,
    user_stack_scoring: bool = True,
) -> int:
    """Run the given function (partial annealing method) in a subprocess and output the produced report."""
    if not with_devel:
        _LOGGER.warning("Development dependencies will not be taken into account - see %s", jl("no_dev"))

    start_time = time.monotonic()

    pid = 0
    if _FORK:
        pid = os.fork()

    if pid == 0:  # Child or no-fork mode.
        return_code = 0
        # We need to re-init logging for the sub-process.
        _LOGGER.debug("Created a child process to compute report")
        try:
            report: Union[DependencyMonkeyReport, Report] = resolver.resolve(
                with_devel=with_devel, user_stack_scoring=user_stack_scoring
            )
            if plot:
                parts = plot.rsplit(".", maxsplit=1)
                file_name = parts[0]
                extension = parts[1] if len(parts) == 2 else "png"

                predictor_history_file = f"{file_name}_{resolver.predictor.__class__.__name__}.{extension}"
                beam_history_file = f"{file_name}_{resolver.beam.__class__.__name__}.{extension}"
                resolver_history_file = f"{file_name}_resolver.{extension}"
                try:
                    figure = resolver.predictor.plot()
                    figure.savefig(predictor_history_file, format=extension)
                except Exception as exc:
                    _LOGGER.exception("Failed to plot predictor history to %r: %s", predictor_history_file, str(exc))
                else:
                    _LOGGER.info("Predictor history saved to %r", predictor_history_file)

                try:
                    figure = resolver.beam.plot()
                    figure.savefig(beam_history_file, format=extension)
                except Exception as exc:
                    _LOGGER.exception("Failed to plot beam history to %r: %s", beam_history_file, str(exc))
                else:
                    _LOGGER.info("Beam history saved to %r", beam_history_file)

                try:
                    figure = resolver.plot()
                    figure.savefig(resolver_history_file, format=extension)
                except Exception as exc:
                    _LOGGER.exception("Failed to plot resolver history to %r: %s", resolver_history_file, str(exc))
                else:
                    _LOGGER.info("Resolver history saved to %r", resolver_history_file)

            result_dict.update(dict(error=False, error_msg=None, report=report.to_dict(verbose=verbose)))
        except UnresolvedDependencies as exc:
            _LOGGER.error(
                "Resolver failed due to unsolved dependencies for packages %s",
                ", ".join(exc.unresolved),
            )
            return_code = 2  # If forked, do not overwrite results by parent process.
            result_dict.update(dict(error=True, error_msg=str(exc), report=exc.to_dict()))
        except CannotProduceStack as exc:
            _LOGGER.error("Resolver did not produce any software stack: %s", str(exc))
            return_code = 2  # If forked, do not overwrite results by parent process.
            result_dict.update(dict(error=True, error_msg=str(exc), report=exc.to_dict()))
        except Exception as exc:
            error_msg = f"The resolution failed as an error was encountered: {str(exc)}"
            _LOGGER.exception(error_msg)
            result_dict.update(
                dict(error=True, error_msg=error_msg, report=dict(ERROR="An error occurred, see logs for more info"))
            )
            return_code = 2

        # Always submit results, even on error.
        print_func(time.monotonic() - start_time, result_dict)

        if _FORK:
            # 1 - error based on user input
            # 2 - error based on system not capable giving recommendations (not enough data) or internal error.
            os._exit(return_code)

        return return_code
    else:  # Parent waits for its child to terminate.
        _LOGGER.debug("Waiting for child process %r", pid)
        _, exit_code = os.waitpid(pid, 0)
        if exit_code != 0:
            _LOGGER.error("Child exited with exit code %r", exit_code)
            if (exit_code & 0xF) == 9:
                err_msg = f"Resolver was killed as allocated memory has been exceeded (OOM) - {jl('oom')}"
            elif os.path.isfile(_LIVENESS_PROBE_KILL_FILE):
                err_msg = f"Resolver was killed as allocated CPU time was exceeded - {jl('cpu_time_exceeded')}"
            else:
                err_msg = (
                    f"Resolution was terminated based on errors encountered; "
                    f"see logs for more info - {jl('error_logs')}"
                )

            _LOGGER.error(err_msg)

            if _FORK and (exit_code >> 8) == 2:
                # Do not overwrite results computed in the forked process.
                return exit_code

            result_dict.update(dict(error=True, error_msg=err_msg, report=None))
            print_func(time.monotonic() - start_time, result_dict)
        else:
            _LOGGER.debug("Subprocess computing report finished successfully")

        # Propagate exit code from child process.
        return exit_code



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2020 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Implementation of predictor used for resolving latest stacks in the state space."""

import logging
import math

import attr
from typing import Optional
from typing import Tuple
from typing import Set
from typing import Any
from typing import List

from thoth.common import get_justification_link as jl

from .hill_climbing import HillClimbing
from ..state import State


_LOGGER = logging.getLogger(__name__)


@attr.s(slots=True)
class ApproximatingLatest(HillClimbing):
    """Implementation of predictor used for resolving latest stacks in the state space.

    This predictor approximates resolution to the latest software stack. The resolution to the latest is
    approximated using continuous resolution with an optional randomness not to get stuck in a "trap"
    if resolution to all latest cannot be satisfied.
    """

    prioritized_packages = attr.ib(type=List[str], default=attr.Factory(list), kw_only=True)
    _hop = attr.ib(type=bool, default=False, init=False)
    _hop_logged = attr.ib(type=bool, default=False, init=False)
    _packages_heated_up = attr.ib(type=Set[str], factory=set, init=False)
    _initial_state = attr.ib(type=Optional[State], default=None, init=False)
    _latest_versions_heat_up = attr.ib(type=Set[Any], factory=set, init=False)

    def set_reward_signal(self, state: State, package_tuple: Tuple[str, str, str], reward: float) -> None:
        """Set hop to True if we did not get resolve any stack with latest."""
        super().set_reward_signal(state, package_tuple, reward)

        if math.isnan(reward):
            self._hop = True
            if not self._hop_logged:
                _LOGGER.warning(
                    "The latest stack couldn't be resolved, performing hops across package versions - see %s",
                    jl("latest_hops"),
                )
                self._hop_logged = True

        if math.isinf(reward):
            self._hop = False

    def pre_run(self) -> None:
        """Initialize local variables before each predictor run per resolver."""
        super().pre_run()
        self._hop = False
        self._hop_logged = False
        self._initial_state = None
        self._packages_heated_up.clear()

    def _heat_up(self) -> Optional[Tuple[State, Tuple[str, str, str]]]:
        """Start heating up phase for the predictor.

        This phase generates new states out of the initial state so that the predictor is not stuck
        in states generated out of the very first state and very first dependency.
        """
        if not self._packages_heated_up:
            self._initial_state = self.context.beam.get(0)

        if self._initial_state is None:
            raise RuntimeError
        # Keeping the initial state as an attribute in this predictor and this
        # check is an optimization thanks to the logic behind resolver's state
        # manipulation - it reuses the initial state for generating new states
        # (see memory optimization) - if the initial state starts to have
        # resolved dependencies, it means there are no more unresolved
        # dependencies to be tracked or the newly cloned state out of the
        # initial state would be the same as the initial state.
        if self._initial_state.resolved_dependencies:
            # End the heat up phase.
            self._initial_state = None
            return None

        for unresolved_dependency in self._initial_state.unresolved_dependencies:
            if unresolved_dependency in self._packages_heated_up:
                continue

            self._packages_heated_up.add(unresolved_dependency)
            unresolved_dependency_tuple = next(
                iter(self._initial_state.unresolved_dependencies[unresolved_dependency].values())
            )

            if self.keep_history:
                self._history.append(
                    (
                        self._initial_state.score,
                        self.context.accepted_final_states_count,
                    )
                )

            return self._initial_state, unresolved_dependency_tuple

        self._initial_state = None
        return None

    def run(self) -> Tuple[State, Tuple[str, str, str]]:
        """Get last state expanded and expand first unresolved dependency."""
        if not self._packages_heated_up or self._initial_state:
            heat_up_result = self._heat_up()
            if heat_up_result is not None:
                return heat_up_result

        state = self.context.beam.get_last()

        if state is None:
            state = self.context.beam.get_random()

        if self.keep_history:
            self._history.append((state.score, self.context.accepted_final_states_count))

        if self._hop:
            for prioritized_package in self.prioritized_packages:
                if prioritized_package in state.unresolved_dependencies:
                    return state, state.get_random_unresolved_dependency(prioritized_package, prefer_recent=True)

            return state, state.get_random_unresolved_dependency(prefer_recent=True)

        for prioritized_package in self.prioritized_packages:
            if prioritized_package in state.unresolved_dependencies:
                return state, state.get_first_unresolved_dependency(prioritized_package)

        return state, state.get_first_unresolved_dependency()



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2019 - 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a sieve that provides only packages available on Operate First Pulp instance."""

from thoth.adviser.context import Context
from thoth.adviser.enums import RecommendationType
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.sieves import NoPulpIndexLabelSieve
from thoth.adviser.sieves._pulp import PULP_URL
from thoth.python import PackageVersion
from thoth.python import Source

from ...base import AdviserUnitTestCase


class TestNoPulpIndexLabelSieve(AdviserUnitTestCase):
    """Test a sieve that provides only packages available on Operate First Pulp instance."""

    UNIT_TESTED = NoPulpIndexLabelSieve

    def test_verify_multiple_should_include(self) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        builder_context = PipelineBuilderContext(
            recommendation_type=RecommendationType.LATEST, labels={"opf-pulp-indexes": "disabled"}
        )
        self.verify_multiple_should_include(builder_context)

    def test_no_include(self, builder_context: PipelineBuilderContext) -> None:
        """Test not including this pipeline unit if the index=no-pulp label is not provided."""
        builder_context.labels.clear()
        self.UNIT_TESTED.should_include(builder_context)

    def test_run(self, context: Context) -> None:
        """Test filtering packages that are not hosted on Operate First Pulp."""
        tf_pypi = PackageVersion(
            name="tensorflow",
            version="==2.7.0",
            index=Source("https://pypi.org/simple"),
            develop=False,
        )

        tf_pulp = PackageVersion(
            name="tensorflow",
            version="==2.7.0",
            index=Source(f"{PULP_URL}simple"),
            develop=False,
        )

        with self.UNIT_TESTED.assigned_context(context):
            sieve = self.UNIT_TESTED()
            assert list(sieve.run(p for p in [tf_pypi, tf_pulp])) == [tf_pypi]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a boot that provides a link to Thoth Search UI."""

import random

from thoth.adviser.boots import ThothSearchBoot
from thoth.adviser.context import Context
from thoth.adviser.pipeline_builder import PipelineBuilderContext
from ..base import AdviserUnitTestCase


class TestThothSearchBoot(AdviserUnitTestCase):
    """Test a boot that provides a link to Thoth Search UI."""

    UNIT_TESTED = ThothSearchBoot

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_run(self, context: Context) -> None:
        """Test propagating a link to Thoth Search UI."""
        assert not context.stack_info

        document_id = f"adviser-foo-bar-{random.randint(0, 1000)}"
        search_url = "https://thoth-station.ninja/some/nice/url/to/search/{document_id}/summary/link"

        unit = self.UNIT_TESTED(document_id=document_id, search_url=search_url)
        with unit.assigned_context(context):
            assert unit.run() is None

        expected_link = search_url.format(document_id=document_id)
        assert context.stack_info == [
            {
                "type": "INFO",
                "message": "Results can be browsed in Thoth search",
                "link": expected_link,
            }
        ]



#!/usr/bin/env python3
# thoth-adviser
# Copyright(C) 2021 Fridolin Pokorny
#
# This program is free software: you can redistribute it and / or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""Test a wrap that provides a link to Thoth Search UI for packages."""

from thoth.adviser.pipeline_builder import PipelineBuilderContext
from thoth.adviser.context import Context
from thoth.adviser.state import State
from thoth.adviser.wraps import ThothSearchPackageWrap

from ..base import AdviserUnitTestCase


class TestThothSearchPackageWrap(AdviserUnitTestCase):
    """Test a wrap that provides a link to Thoth Search UI for packages."""

    UNIT_TESTED = ThothSearchPackageWrap

    def test_verify_multiple_should_include(self, builder_context: PipelineBuilderContext) -> None:
        """Verify multiple should_include calls do not loop endlessly."""
        self.verify_multiple_should_include(builder_context)

    def test_run(self, context: Context, state: State) -> None:
        """Test propagating a link to Thoth Search UI."""
        search_package_url = (
            "https://thoth-station.ninja/some/nice/url/pkg/{package_name}/ver/{package_version}"
            "/idx/{index_url}/os/{os_name}/os_ver/{os_version}/py/{python_version}/summary"
        )

        context.project.runtime_environment.operating_system.name = "fedora"
        context.project.runtime_environment.operating_system.version = "35"
        context.project.runtime_environment.python_version = "3.10"

        state.justification.clear()
        state.resolved_dependencies.clear()
        state.add_resolved_dependency(("flask", "2.0.2", "https://pypi.org/simple"))
        state.add_resolved_dependency(("thamos", "1.21.1", "https://thoth-station.ninja/simple"))

        unit = self.UNIT_TESTED(search_package_url=search_package_url)
        with unit.assigned_context(context):
            assert unit.run(state) is None

        assert state.justification == [
            {
                "link": "https://thoth-station.ninja/some/nice/url/pkg/flask/ver"
                "/2.0.2/idx/https%3A%2F%2Fpypi.org%2Fsimple/os/fedora/os_ver/35/py/3.10/summary",
                "message": "Browse Thoth Search UI for 'flask==2.0.2'",
                "package_name": "flask",
                "type": "INFO",
            },
            {
                "link": "https://thoth-station.ninja/some/nice/url/pkg/thamos/ver"
                "/1.21.1/idx/https%3A%2F%2Fthoth-station.ninja%2Fsimple/os/fedora/os_ver/35/py/3.10/summary",
                "message": "Browse Thoth Search UI for 'thamos==1.21.1'",
                "package_name": "thamos",
                "type": "INFO",
            },
        ]
